from typing import Optional, List, Any
import logging

from overmind.clients.overmind_client import OvermindClient

try:
    import opentelemetry.instrumentation.google_generativeai as __google_instrumentation

    __google_instrumentation.GoogleGenerativeAiInstrumentor().instrument()
    from google import genai
except ImportError:
    raise ImportError(
        "google-genai is not installed. Please install it with `pip install google-genai`."
    )

logger = logging.getLogger(__name__)


class Client(genai.Client):
    def __init__(
        self,
        *,
        overmind_api_key: Optional[str] = None,
        overmind_base_url: Optional[str] = None,
        input_policies: List[str] = [],
        output_policies: List[str] = [],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.overmind_client = OvermindClient(overmind_api_key, overmind_base_url)
        self.global_input_policies = input_policies
        self.global_output_policies = output_policies
        # Wrap generate_content for policy interception
        self._original_generate_content = self.models.generate_content
        self.models.generate_content = self._wrapped_generate_content

    def _run_input_policies(self, contents):
        """Run input policies on the contents, modifying text in place."""
        if isinstance(contents, str):
            for policy in self.global_input_policies:
                res = self.overmind_client.run_layer(
                    policy=policy, input_data=contents
                )
                contents = res.processed_data or contents
            return contents

        if isinstance(contents, list) and contents:
            last = contents[-1]
            if isinstance(last, dict):
                parts = last.get("parts", [])
                for part in parts:
                    if isinstance(part, dict) and "text" in part:
                        text = part["text"]
                        for policy in self.global_input_policies:
                            res = self.overmind_client.run_layer(
                                policy=policy, input_data=text
                            )
                            part["text"] = res.processed_data or text

        return contents

    def _extract_question(self, contents) -> str:
        """Extract the question text from contents for output policy context."""
        if isinstance(contents, str):
            return contents
        if isinstance(contents, list) and contents:
            last = contents[-1]
            if isinstance(last, str):
                return last
            if isinstance(last, dict):
                parts = last.get("parts", [])
                texts = [
                    p.get("text", "")
                    for p in parts
                    if isinstance(p, dict) and "text" in p
                ]
                return " ".join(texts)
        return str(contents)

    def _wrapped_generate_content(self, **kwargs):
        contents = kwargs.get("contents")

        # Run input policies
        if self.global_input_policies and contents:
            kwargs["contents"] = self._run_input_policies(contents)

        question = self._extract_question(kwargs.get("contents", contents))

        # Call original generate_content
        response = self._original_generate_content(**kwargs)

        # Run output policies
        if self.global_output_policies:
            try:
                response_text = response.text
            except Exception:
                response_text = None

            if response_text:
                for policy in self.global_output_policies:
                    self.overmind_client.run_layer(
                        policy=policy,
                        input_data=response_text,
                        question=question,
                    )

        return response



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, List, Any
import logging

from httpx import Request, Response, URL

from overmind.clients.overmind_client import OvermindClient

try:
    import opentelemetry.instrumentation.anthropic as __anthropic_instrumentation

    __anthropic_instrumentation.AnthropicInstrumentor().instrument()
    import anthropic as __anthropic
except ImportError:
    raise ImportError(
        "anthropic is not installed. Please install it with `pip install anthropic`."
    )

logger = logging.getLogger(__name__)


class Anthropic(__anthropic.Anthropic):
    def __init__(
        self,
        *,
        overmind_api_key: Optional[str] = None,
        overmind_base_url: Optional[str] = None,
        input_policies: List[str] = [],
        output_policies: List[str] = [],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.overmind_client = OvermindClient(overmind_api_key, overmind_base_url)
        self._client.build_request = self.build_request_decorator(
            self._client.build_request
        )
        self.global_input_policies = input_policies
        self.global_output_policies = output_policies
        self._client.event_hooks = {
            "response": [self.run_output_policies],
        }

    def _get_handler(self, path: str) -> Optional["BaseHandler"]:
        if path == messagesHandler.path:
            logger.debug(f"using messagesHandler for path: {path}")
            return messagesHandler(self.overmind_client)
        return None

    def build_request_decorator(self, func):
        def wrapper(*args, **kwargs) -> Request:
            json_content = kwargs.get("json")
            if not json_content:
                return func(*args, **kwargs)

            input_policies = (
                json_content.pop("input_policies", None) or self.global_input_policies
            )
            output_policies = (
                json_content.pop("output_policies", None)
                or self.global_output_policies
            )

            url: URL = kwargs.get("url")
            handler = self._get_handler(url.path)

            if handler and not json_content.get("stream", False):
                kwargs = handler.pre_request(
                    kwargs=kwargs,
                    json_content=json_content,
                    input_policies=input_policies,
                )
            else:
                kwargs.update(json=json_content)

            request: Request = func(*args, **kwargs)
            setattr(request, "input_policies", input_policies)
            setattr(request, "output_policies", output_policies)
            if handler:
                handler.post_request(request=request, kwargs=kwargs)

            return request

        return wrapper

    def run_output_policies(self, response: Response):
        if response.status_code != 200:
            return

        if response.headers.get("Content-Type") != "application/json":
            return

        output_policies = getattr(response._request, "output_policies", [])
        logger.info(f"output_policies={output_policies} in response")

        response.read()
        json_output = response.json()
        handler = self._get_handler(response._request.url.path)

        if handler:
            handler.post_response(
                response=response,
                json_output=json_output,
                output_policies=output_policies,
            )


@dataclass
class BaseHandler(ABC):
    overmind_client: OvermindClient
    path: str = None

    def _run_input_policies_messages(
        self, inputs: list[str], input_policies: list[Any]
    ):
        for policy in input_policies:
            res = self.overmind_client.run_layer(policy=policy, input_data=inputs[-1])
            inputs[-1] = res.processed_data or inputs[-1]
        return inputs

    @abstractmethod
    def pre_request(
        self, kwargs: dict, json_content: dict, input_policies: list[Any]
    ): ...

    @abstractmethod
    def post_request(self, request: Request, kwargs: dict): ...

    @abstractmethod
    def post_response(
        self, response: Response, json_output: dict, output_policies: list[Any]
    ): ...


class messagesHandler(BaseHandler):
    path = "/v1/messages"

    def pre_request(self, kwargs: dict, json_content: dict, input_policies: list[Any]):
        messages = json_content.get("messages", [])
        if messages and input_policies:
            last_message = messages[-1]
            content = last_message.get("content", "")
            # Anthropic supports string content or list of content blocks
            if isinstance(content, str):
                modified = self._run_input_policies_messages(
                    inputs=[content], input_policies=input_policies
                )[-1]
                messages[-1] = {**last_message, "content": modified}
            elif isinstance(content, list):
                for block in content:
                    if isinstance(block, dict) and block.get("type") == "text":
                        modified = self._run_input_policies_messages(
                            inputs=[block["text"]], input_policies=input_policies
                        )[-1]
                        block["text"] = modified
            json_content["messages"] = messages
        kwargs.update(json=json_content)
        return kwargs

    def post_request(self, request: Request, kwargs: dict):
        messages = kwargs.get("json", {}).get("messages", [])
        question = ""
        if messages:
            content = messages[-1].get("content", "")
            if isinstance(content, str):
                question = content
            elif isinstance(content, list):
                question = " ".join(
                    b.get("text", "")
                    for b in content
                    if isinstance(b, dict) and b.get("type") == "text"
                )
        setattr(request, "kwargs", {"question": question})

    def post_response(
        self, response: Response, json_output: dict, output_policies: list[Any]
    ):
        # Anthropic response format: {"content": [{"type": "text", "text": "..."}]}
        content_blocks = json_output.get("content", [])
        text_content = " ".join(
            block.get("text", "")
            for block in content_blocks
            if isinstance(block, dict) and block.get("type") == "text"
        )

        if not text_content:
            return

        for policy in output_policies:
            self.overmind_client.run_layer(
                policy=policy,
                input_data=text_content,
                **getattr(response._request, "kwargs", {}),
            )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, Callable, Any, List
import logging

from httpx import Request, Response, URL

from overmind.clients.overmind_client import OvermindClient

try:
    import openai as __openai
except ImportError:
    raise ImportError(
        "openai is not installed. Please install it with `pip install openai`."
    )

logger = logging.getLogger(__name__)


class OpenAI(__openai.OpenAI):
    # init_overmind
    def __init__(
        self,
        *,
        overmind_api_key: Optional[str] = None,
        # internal client will use OVERMIND_API_KEY env var
        overmind_base_url: Optional[str] = None,
        # internal client will use OVERMIND_API_URL env var
        # a set of global policies that will be applied to all requests
        input_policies: List[str] = [],
        # a set of global policies that will be applied to all responses
        output_policies: List[str] = [],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.overmind_client = OvermindClient(overmind_api_key, overmind_base_url)
        self._client.build_request = self.build_request_decorator(
            self._client.build_request
        )
        self.global_input_policies = input_policies
        self.global_output_policies = output_policies
        self._client.event_hooks = {
            "response": [self.run_output_policies],
        }

    def _get_handler(self, path: str) -> Optional["BaseHandler"]:
        if path == responsesHandler.path:
            logger.debug(f"using responsesHandler for path: {path}")
            return responsesHandler(self.overmind_client)
        if path == chatCompletionHandler.path:
            logger.debug(f"using chatCompletionHandler for path: {path}")
            return chatCompletionHandler(self.overmind_client)
        if path == audioSpeechHandler.path:
            logger.debug(f"using audioSpeechHandler for path: {path}")
            return audioSpeechHandler(self.overmind_client)
        if path == responsesCompactHandler.path:
            logger.debug(f"using responsesCompactHandler for path: {path}")
            return responsesCompactHandler(self.overmind_client)

    def build_request_decorator(
        self, func: Callable[[Any], Request]
    ) -> Callable[[Any], Request]:
        # TODO: not the best idea to add to extra_body
        def wrapper(*args, **kwargs) -> Request:
            # TODO: need to handle all, content, body, and data, and files
            json_content = kwargs.get("json")
            if not json_content:
                return func(*args, **kwargs)
            # role = input policy will change the user input
            # output policy will  not change the response but just flags it to the user
            input_policies = (
                json_content.pop("input_policies", None) or self.global_input_policies
            )
            output_policies = (
                json_content.pop("output_policies", None) or self.global_output_policies
            )

            url: URL = kwargs.get("url")
            handler = self._get_handler(url.path)
            # TODO: add stram handler here
            if handler and not json_content.get("stream", False):
                kwargs = handler.pre_request(
                    kwargs=kwargs,
                    json_content=json_content,
                    input_policies=input_policies,
                )
            else:
                kwargs.update(json=json_content)
            request: Request = func(*args, **kwargs)
            setattr(request, "input_policies", input_policies)
            setattr(request, "output_policies", output_policies)
            if handler:
                handler.post_request(request=request, kwargs=kwargs)

            # run input policies here, modify the request content as needed
            return request

        return wrapper

    def run_output_policies(self, response: Response):
        def summary():
            raise NotImplementedError("method not implemented yet")

        setattr(response, "summary", summary)
        # return
        if response.status_code != 200:
            return

        # allow audio responses/ non text responses
        if response.headers.get("Content-Type") != "application/json":
            return

        output_policies = getattr(response._request, "output_policies", [])
        logger.info(f"output_policies={output_policies} in response")

        response.read()  # read the response content to get the actual content
        # TODO: need a few checks here for all kinds of payloads
        json_output = response.json()
        handler = self._get_handler(response._request.url.path)

        if handler:
            handler.post_response(
                response=response,
                json_output=json_output,
                output_policies=output_policies,
            )


@dataclass
class BaseHandler(ABC):
    overmind_client: OvermindClient
    path: str = None

    def run_input_policies(self, messages: list[dict], input_policies: List[Any]):
        # ideally should run it for all items in the array of messages messages
        content_modified = self._run_input_policies_messages(
            [messages[-1]["content"]], input_policies
        )[-1]

        messages[-1]["content"] = content_modified
        return messages

    def _run_input_policies_messages(
        self, inputs: list[str], input_policies: list[Any]
    ):
        for policy in input_policies:
            res = self.overmind_client.run_layer(policy=policy, input_data=inputs[-1])
            inputs[-1] = (
                res.processed_data or inputs[-1]
            )  # if the policy returns None, keep the original input
        return inputs

    @abstractmethod
    def pre_request(
        self,
        kwargs: dict,
        json_content: dict,
        input_policies: list[Any],
    ): ...

    @abstractmethod
    def post_request(self, request: Request, kwargs: dict): ...

    @abstractmethod
    def post_response(
        self, response: Response, json_output: dict, output_policies: list[Any]
    ): ...


class chatCompletionHandler(BaseHandler):
    path = "/v1/chat/completions"

    def pre_request(self, kwargs: dict, json_content: dict, input_policies: list[Any]):
        kwargs.update(json=self._pre_json(json_content, input_policies))
        return kwargs

    def _pre_json(self, json_content: dict, input_policies: list[Any]):
        messages = json_content.pop("messages", None)
        original_message = messages[-1]["content"]
        messages[-1] = {
            "role": messages[-1]["role"],
            "content": self._run_input_policies_messages(
                inputs=[original_message], input_policies=input_policies
            )[-1],
        }
        json_content["messages"] = messages
        return json_content

    def _pre_file(self, file: dict, input_policies: list[Any]): ...

    def post_request(self, request: Request, kwargs: dict):
        setattr(
            request,
            "kwargs",
            {"question": kwargs["json"]["messages"][-1]["content"]},
        )

    def post_response(
        self, response: Response, json_output: dict, output_policies: list[Any]
    ):
        content = json_output["choices"][0]["message"].get("content", "")
        tool_calls = json_output["choices"][0]["message"].get("tool_calls", [])

        if not content:
            return
        for policy in output_policies:
            self.overmind_client.run_layer(
                policy=policy,
                input_data=content,
                **getattr(response._request, "kwargs", {}),
            )


class responsesHandler(BaseHandler):
    path = "/v1/responses"

    def pre_request(self, kwargs: dict, json_content: dict, input_policies: list[Any]):
        modified_input = self._run_input_policies_messages(
            inputs=[json_content.pop("input", None)],
            input_policies=input_policies,
        )[-1]
        json_content.update({"input": modified_input})
        kwargs.update(json=json_content)
        return kwargs

    def post_request(self, request: Request, kwargs: dict):
        setattr(
            request,
            "kwargs",
            {"question": kwargs["json"]["input"]},
        )

    def post_response(
        self, response: Response, json_output: dict, output_policies: list[Any]
    ):
        pass


class doNothingHandler(BaseHandler):
    path = None

    def pre_request(self, kwargs: dict, json_content: dict, input_policies: list[Any]):
        return kwargs

    def post_request(self, request: Request, kwargs: dict):
        return

    def post_response(
        self, response: Response, json_output: dict, output_policies: list[Any]
    ):
        return


class audioSpeechHandler(doNothingHandler):
    path = "/v1/audio/speech"


class responsesCompactHandler(doNothingHandler):
    path = "/v1/responses/compact"


class AsyncOpenAI(__openai.AsyncOpenAI):
    """
    Async Overmind-wrapped OpenAI client.

    Telemetry (OpenTelemetry traces) is fully captured via the OvermindClient
    initialisation which instruments the openai library.

    Policy support (input_policies / output_policies) is not yet wired for
    async – the parameters are accepted and stored so the interface is
    forward-compatible with the sync OpenAI client.
    """

    def __init__(
        self,
        *,
        overmind_api_key: Optional[str] = None,
        overmind_base_url: Optional[str] = None,
        input_policies: List[str] = [],
        output_policies: List[str] = [],
        **kwargs,
    ):
        super().__init__(**kwargs)
        # This triggers overmind.init() which sets up OTel tracing +
        # instruments the openai module (both sync and async paths).
        self.overmind_client = OvermindClient(overmind_api_key, overmind_base_url)
        self.global_input_policies = input_policies
        self.global_output_policies = output_policies



import json
import pydantic
import re
import threading
import time
from typing import Any, Optional, Union

from openai import AsyncStream, Stream
from openai._legacy_response import LegacyAPIResponse
from opentelemetry import context as context_api
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY
from opentelemetry.semconv._incubating.attributes import (
    gen_ai_attributes as GenAIAttributes,
    openai_attributes as OpenAIAttributes,
)
from opentelemetry.semconv.attributes.error_attributes import ERROR_TYPE
from opentelemetry.semconv_ai import SpanAttributes
from opentelemetry.trace import SpanKind, Span, StatusCode, Tracer
from typing_extensions import NotRequired
from wrapt import ObjectProxy

from ..shared import (
    _extract_model_name_from_provider_format,
    _set_request_attributes,
    _set_span_attribute,
    model_as_dict,
)
from ..utils import (
    _with_tracer_wrapper,
    dont_throw,
    should_send_prompts,
)


def _get_openai_sentinel_types() -> tuple:
    """Dynamically discover OpenAI sentinel types available in this SDK version.

    OpenAI SDK uses sentinel objects (NOT_GIVEN, Omit) for unset optional parameters.
    These types may not exist in older SDK versions, so we discover them at runtime.
    """
    sentinel_types = []
    try:
        from openai import NotGiven
        sentinel_types.append(NotGiven)
    except ImportError:
        pass
    try:
        from openai import Omit
        sentinel_types.append(Omit)
    except ImportError:
        pass
    return tuple(sentinel_types)


# Tuple of OpenAI sentinel types for isinstance() checks (empty if none available)
_OPENAI_SENTINEL_TYPES: tuple = _get_openai_sentinel_types()

# Conditional imports for backward compatibility
try:
    from openai.types.responses import (
        FunctionToolParam,
        Response,
        ResponseInputItemParam,
        ResponseInputParam,
        ResponseOutputItem,
        ResponseUsage,
        ToolParam,
    )
    from openai.types.responses.response_output_message_param import (
        ResponseOutputMessageParam,
    )
    RESPONSES_AVAILABLE = True
except ImportError:
    # Fallback types for older OpenAI SDK versions
    from typing import Dict, List

    # Create basic fallback types
    FunctionToolParam = Dict[str, Any]
    Response = Any
    ResponseInputItemParam = Dict[str, Any]
    ResponseInputParam = Union[str, List[Dict[str, Any]]]
    ResponseOutputItem = Dict[str, Any]
    ResponseUsage = Dict[str, Any]
    ToolParam = Dict[str, Any]
    ResponseOutputMessageParam = Dict[str, Any]
    RESPONSES_AVAILABLE = False

SPAN_NAME = "openai.response"


def _sanitize_sentinel_values(kwargs: dict) -> dict:
    """Remove OpenAI sentinel values (NOT_GIVEN, Omit) from kwargs.

    OpenAI SDK uses sentinel objects for unset optional parameters.
    These don't have dict methods like .get(), causing errors when
    code chains calls like kwargs.get("reasoning", {}).get("summary").

    This removes sentinel values so the default (e.g., {}) is used instead
    when calling .get() on the sanitized dict.

    If no sentinel types are available (older SDK), returns kwargs unchanged.
    """
    if not _OPENAI_SENTINEL_TYPES:
        return kwargs
    return {k: v for k, v in kwargs.items()
            if not isinstance(v, _OPENAI_SENTINEL_TYPES)}


def prepare_input_param(input_param: ResponseInputItemParam) -> ResponseInputItemParam:
    """
    Looks like OpenAI API infers the type "message" if the shape is correct,
    but type is not specified.
    It is marked as required on the message types. We add this to our
    traced data to make it work.
    """
    try:
        d = model_as_dict(input_param)
        if "type" not in d:
            d["type"] = "message"
        if RESPONSES_AVAILABLE:
            return ResponseInputItemParam(**d)
        else:
            return d
    except Exception:
        return input_param


def process_input(inp: ResponseInputParam) -> ResponseInputParam:
    if not isinstance(inp, list):
        return inp
    return [prepare_input_param(item) for item in inp]


def is_validator_iterator(content):
    """
    Some OpenAI objects contain fields typed as Iterable, which pydantic
    internally converts to a ValidatorIterator, and they cannot be trivially
    serialized without consuming the iterator to, for example, a list.

    See: https://github.com/pydantic/pydantic/issues/9541#issuecomment-2189045051
    """
    return re.search(r"pydantic.*ValidatorIterator'>$", str(type(content)))


# OpenAI API accepts output messages without an ID in its inputs, but
# the ID is marked as required in the output type.
if RESPONSES_AVAILABLE:
    class ResponseOutputMessageParamWithoutId(ResponseOutputMessageParam):
        id: NotRequired[str]
else:
    # Fallback for older SDK versions
    ResponseOutputMessageParamWithoutId = dict


class TracedData(pydantic.BaseModel):
    start_time: float  # time.time_ns()
    response_id: str
    # actually Union[str, list[Union[ResponseInputItemParam, ResponseOutputMessageParamWithoutId]]],
    # but this only works properly in Python 3.10+ / newer pydantic
    input: Any
    # system message
    instructions: Optional[str] = pydantic.Field(default=None)
    # TODO: remove Any with newer Python / pydantic
    tools: Optional[list[Union[Any, ToolParam]]] = pydantic.Field(default=None)
    output_blocks: Optional[dict[str, ResponseOutputItem]] = pydantic.Field(
        default=None
    )
    usage: Optional[ResponseUsage] = pydantic.Field(default=None)
    output_text: Optional[str] = pydantic.Field(default=None)
    request_model: Optional[str] = pydantic.Field(default=None)
    response_model: Optional[str] = pydantic.Field(default=None)

    # Reasoning attributes
    request_reasoning_summary: Optional[str] = pydantic.Field(default=None)
    request_reasoning_effort: Optional[str] = pydantic.Field(default=None)
    response_reasoning_effort: Optional[str] = pydantic.Field(default=None)

    # OpenAI service tier
    request_service_tier: Optional[str] = pydantic.Field(default=None)
    response_service_tier: Optional[str] = pydantic.Field(default=None)

    # Trace context - to maintain trace continuity across async operations
    trace_context: Any = pydantic.Field(default=None)

    class Config:
        arbitrary_types_allowed = True


responses: dict[str, TracedData] = {}


def parse_response(response: Union[LegacyAPIResponse, Response]) -> Response:
    if isinstance(response, LegacyAPIResponse):
        return response.parse()
    return response


def get_tools_from_kwargs(kwargs: dict) -> list[ToolParam]:
    tools_input = kwargs.get("tools", [])
    # Handle case where tools key exists but value is None
    # (e.g., when wrappers like openai-guardrails pass tools=None)
    if tools_input is None:
        tools_input = []
    tools = []

    for tool in tools_input:
        if tool.get("type") == "function":
            if RESPONSES_AVAILABLE:
                tools.append(FunctionToolParam(**tool))
            else:
                tools.append(tool)

    return tools


def process_content_block(
    block: dict[str, Any],
) -> dict[str, Any]:
    # TODO: keep the original type once backend supports it
    if block.get("type") in ["text", "input_text", "output_text"]:
        return {"type": "text", "text": block.get("text")}
    elif block.get("type") in ["image", "input_image", "output_image"]:
        return {
            "type": "image",
            "image_url": block.get("image_url"),
            "detail": block.get("detail"),
            "file_id": block.get("file_id"),
        }
    elif block.get("type") in ["file", "input_file", "output_file"]:
        return {
            "type": "file",
            "file_id": block.get("file_id"),
            "filename": block.get("filename"),
            "file_data": block.get("file_data"),
        }
    return block


@dont_throw
def prepare_kwargs_for_shared_attributes(kwargs):
    """
    Prepare kwargs for the shared _set_request_attributes function.
    Maps responses API specific parameters to the common format.
    """
    prepared_kwargs = kwargs.copy()

    # Map max_output_tokens to max_tokens for the shared function
    if "max_output_tokens" in kwargs:
        prepared_kwargs["max_tokens"] = kwargs["max_output_tokens"]

    return prepared_kwargs


def set_data_attributes(traced_response: TracedData, span: Span):
    _set_span_attribute(span, GenAIAttributes.GEN_AI_RESPONSE_ID, traced_response.response_id)

    response_model = _extract_model_name_from_provider_format(traced_response.response_model)
    _set_span_attribute(span, GenAIAttributes.GEN_AI_RESPONSE_MODEL, response_model)

    _set_span_attribute(span, OpenAIAttributes.OPENAI_RESPONSE_SERVICE_TIER, traced_response.response_service_tier)
    if usage := traced_response.usage:
        _set_span_attribute(span, GenAIAttributes.GEN_AI_USAGE_INPUT_TOKENS, usage.input_tokens)
        _set_span_attribute(span, GenAIAttributes.GEN_AI_USAGE_OUTPUT_TOKENS, usage.output_tokens)
        _set_span_attribute(
            span, SpanAttributes.LLM_USAGE_TOTAL_TOKENS, usage.total_tokens
        )
        if usage.input_tokens_details:
            _set_span_attribute(
                span,
                SpanAttributes.LLM_USAGE_CACHE_READ_INPUT_TOKENS,
                usage.input_tokens_details.cached_tokens,
            )

        reasoning_tokens = None
        tokens_details = (
            usage.get("output_tokens_details") if isinstance(usage, dict)
            else getattr(usage, "output_tokens_details", None)
        )

        if tokens_details:
            reasoning_tokens = (
                tokens_details.get("reasoning_tokens", None) if isinstance(tokens_details, dict)
                else getattr(tokens_details, "reasoning_tokens", None)
            )

        _set_span_attribute(
            span,
            SpanAttributes.LLM_USAGE_REASONING_TOKENS,
            reasoning_tokens or 0,
        )

    _set_span_attribute(
        span,
        f"{SpanAttributes.LLM_REQUEST_REASONING_SUMMARY}",
        traced_response.request_reasoning_summary or (),
    )
    _set_span_attribute(
        span,
        f"{SpanAttributes.LLM_REQUEST_REASONING_EFFORT}",
        traced_response.request_reasoning_effort or (),
    )
    _set_span_attribute(
        span,
        f"{SpanAttributes.LLM_RESPONSE_REASONING_EFFORT}",
        traced_response.response_reasoning_effort or (),
    )

    if should_send_prompts():
        prompt_index = 0
        if traced_response.tools:
            for i, tool_param in enumerate(traced_response.tools):
                tool_dict = model_as_dict(tool_param)
                description = tool_dict.get("description")
                parameters = tool_dict.get("parameters")
                name = tool_dict.get("name")
                if parameters is None:
                    continue
                _set_span_attribute(
                    span,
                    f"{SpanAttributes.LLM_REQUEST_FUNCTIONS}.{i}.description",
                    description,
                )
                _set_span_attribute(
                    span,
                    f"{SpanAttributes.LLM_REQUEST_FUNCTIONS}.{i}.parameters",
                    json.dumps(parameters),
                )
                _set_span_attribute(
                    span,
                    f"{SpanAttributes.LLM_REQUEST_FUNCTIONS}.{i}.name",
                    name,
                )
        if traced_response.instructions:
            _set_span_attribute(
                span,
                f"{GenAIAttributes.GEN_AI_PROMPT}.{prompt_index}.content",
                traced_response.instructions,
            )
            _set_span_attribute(span, f"{GenAIAttributes.GEN_AI_PROMPT}.{prompt_index}.role", "system")
            prompt_index += 1

        if isinstance(traced_response.input, str):
            _set_span_attribute(
                span, f"{GenAIAttributes.GEN_AI_PROMPT}.{prompt_index}.content", traced_response.input
            )
            _set_span_attribute(span, f"{GenAIAttributes.GEN_AI_PROMPT}.{prompt_index}.role", "user")
            prompt_index += 1
        else:
            for block in traced_response.input:
                block_dict = model_as_dict(block)
                if block_dict.get("type", "message") == "message":
                    content = block_dict.get("content")
                    if is_validator_iterator(content):
                        # we're after the actual call here, so we can consume the iterator
                        content = [process_content_block(block) for block in content]
                    try:
                        stringified_content = (
                            content if isinstance(content, str) else json.dumps(content)
                        )
                    except Exception:
                        stringified_content = (
                            str(content) if content is not None else ""
                        )
                    _set_span_attribute(
                        span,
                        f"{GenAIAttributes.GEN_AI_PROMPT}.{prompt_index}.content",
                        stringified_content,
                    )
                    _set_span_attribute(
                        span,
                        f"{GenAIAttributes.GEN_AI_PROMPT}.{prompt_index}.role",
                        block_dict.get("role"),
                    )
                    prompt_index += 1
                elif block_dict.get("type") == "computer_call_output":
                    _set_span_attribute(
                        span, f"{GenAIAttributes.GEN_AI_PROMPT}.{prompt_index}.role", "computer-call"
                    )
                    output_image_url = block_dict.get("output", {}).get("image_url")
                    if output_image_url:
                        _set_span_attribute(
                            span,
                            f"{GenAIAttributes.GEN_AI_PROMPT}.{prompt_index}.content",
                            json.dumps(
                                [
                                    {
                                        "type": "image_url",
                                        "image_url": {"url": output_image_url},
                                    }
                                ]
                            ),
                        )
                    prompt_index += 1
                elif block_dict.get("type") == "computer_call":
                    _set_span_attribute(
                        span, f"{GenAIAttributes.GEN_AI_PROMPT}.{prompt_index}.role", "assistant"
                    )
                    call_content = {}
                    if block_dict.get("id"):
                        call_content["id"] = block_dict.get("id")
                    if block_dict.get("call_id"):
                        call_content["call_id"] = block_dict.get("call_id")
                    if block_dict.get("action"):
                        call_content["action"] = block_dict.get("action")
                    _set_span_attribute(
                        span,
                        f"{GenAIAttributes.GEN_AI_PROMPT}.{prompt_index}.content",
                        json.dumps(call_content),
                    )
                    prompt_index += 1
                # TODO: handle other block types

        _set_span_attribute(span, f"{GenAIAttributes.GEN_AI_COMPLETION}.0.role", "assistant")
        if traced_response.output_text:
            _set_span_attribute(
                span, f"{GenAIAttributes.GEN_AI_COMPLETION}.0.content", traced_response.output_text
            )
        tool_call_index = 0
        for block in traced_response.output_blocks.values():
            block_dict = model_as_dict(block)
            if block_dict.get("type") == "message":
                # either a refusal or handled in output_text above
                continue
            if block_dict.get("type") == "function_call":
                _set_span_attribute(
                    span,
                    f"{GenAIAttributes.GEN_AI_COMPLETION}.0.tool_calls.{tool_call_index}.id",
                    block_dict.get("id"),
                )
                _set_span_attribute(
                    span,
                    f"{GenAIAttributes.GEN_AI_COMPLETION}.0.tool_calls.{tool_call_index}.name",
                    block_dict.get("name"),
                )
                _set_span_attribute(
                    span,
                    f"{GenAIAttributes.GEN_AI_COMPLETION}.0.tool_calls.{tool_call_index}.arguments",
                    block_dict.get("arguments"),
                )
                tool_call_index += 1
            elif block_dict.get("type") == "file_search_call":
                _set_span_attribute(
                    span,
                    f"{GenAIAttributes.GEN_AI_COMPLETION}.0.tool_calls.{tool_call_index}.id",
                    block_dict.get("id"),
                )
                _set_span_attribute(
                    span,
                    f"{GenAIAttributes.GEN_AI_COMPLETION}.0.tool_calls.{tool_call_index}.name",
                    "file_search_call",
                )
                tool_call_index += 1
            elif block_dict.get("type") == "web_search_call":
                _set_span_attribute(
                    span,
                    f"{GenAIAttributes.GEN_AI_COMPLETION}.0.tool_calls.{tool_call_index}.id",
                    block_dict.get("id"),
                )
                _set_span_attribute(
                    span,
                    f"{GenAIAttributes.GEN_AI_COMPLETION}.0.tool_calls.{tool_call_index}.name",
                    "web_search_call",
                )
                tool_call_index += 1
            elif block_dict.get("type") == "computer_call":
                _set_span_attribute(
                    span,
                    f"{GenAIAttributes.GEN_AI_COMPLETION}.0.tool_calls.{tool_call_index}.id",
                    block_dict.get("call_id"),
                )
                _set_span_attribute(
                    span,
                    f"{GenAIAttributes.GEN_AI_COMPLETION}.0.tool_calls.{tool_call_index}.name",
                    "computer_call",
                )
                _set_span_attribute(
                    span,
                    f"{GenAIAttributes.GEN_AI_COMPLETION}.0.tool_calls.{tool_call_index}.arguments",
                    json.dumps(block_dict.get("action")),
                )
                tool_call_index += 1
            elif block_dict.get("type") == "reasoning":
                reasoning_summary = block_dict.get("summary")
                if reasoning_summary is not None and reasoning_summary != []:
                    if isinstance(reasoning_summary, (dict, list)):
                        reasoning_value = json.dumps(reasoning_summary)
                    else:
                        reasoning_value = reasoning_summary
                    _set_span_attribute(
                        span, f"{GenAIAttributes.GEN_AI_COMPLETION}.0.reasoning", reasoning_value
                    )
            # TODO: handle other block types, in particular other calls


@dont_throw
@_with_tracer_wrapper
def responses_get_or_create_wrapper(tracer: Tracer, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)
    start_time = time.time_ns()

    # Remove OpenAI sentinel values (NOT_GIVEN, Omit) to allow chained .get() calls
    non_sentinel_kwargs = _sanitize_sentinel_values(kwargs)

    try:
        response = wrapped(*args, **kwargs)
        if isinstance(response, Stream):
            # Capture current trace context to maintain trace continuity
            ctx = context_api.get_current()
            span = tracer.start_span(
                SPAN_NAME,
                kind=SpanKind.CLIENT,
                start_time=start_time,
                context=ctx,
            )
            _set_request_attributes(span, prepare_kwargs_for_shared_attributes(non_sentinel_kwargs), instance)

            return ResponseStream(
                span=span,
                response=response,
                start_time=start_time,
                request_kwargs=non_sentinel_kwargs,
                tracer=tracer,
            )
    except Exception as e:
        response_id = non_sentinel_kwargs.get("response_id")
        existing_data = {}
        if response_id and response_id in responses:
            existing_data = responses[response_id].model_dump()
        try:
            traced_data = TracedData(
                start_time=existing_data.get("start_time", start_time),
                response_id=response_id or "",
                input=process_input(
                    non_sentinel_kwargs.get("input", existing_data.get("input", []))
                ),
                instructions=non_sentinel_kwargs.get(
                    "instructions", existing_data.get("instructions")
                ),
                tools=get_tools_from_kwargs(non_sentinel_kwargs) or existing_data.get("tools", []),
                output_blocks=existing_data.get("output_blocks", {}),
                usage=existing_data.get("usage"),
                output_text=non_sentinel_kwargs.get(
                    "output_text", existing_data.get("output_text", "")
                ),
                request_model=non_sentinel_kwargs.get(
                    "model", existing_data.get("request_model", "")
                ),
                response_model=existing_data.get("response_model", ""),
                # Reasoning attributes
                request_reasoning_summary=(
                    non_sentinel_kwargs.get("reasoning", {}).get(
                        "summary", existing_data.get("request_reasoning_summary")
                    )
                ),
                request_reasoning_effort=(
                    non_sentinel_kwargs.get("reasoning", {}).get(
                        "effort", existing_data.get("request_reasoning_effort")
                    )
                ),
                response_reasoning_effort=non_sentinel_kwargs.get("reasoning", {}).get("effort"),
                request_service_tier=non_sentinel_kwargs.get("service_tier"),
                response_service_tier=existing_data.get("response_service_tier"),
                # Capture trace context to maintain continuity
                trace_context=existing_data.get("trace_context", context_api.get_current()),
            )
        except Exception:
            traced_data = None

        # Restore the original trace context to maintain trace continuity
        ctx = (traced_data.trace_context if traced_data and traced_data.trace_context
               else context_api.get_current())
        span = tracer.start_span(
            SPAN_NAME,
            kind=SpanKind.CLIENT,
            start_time=(
                start_time if traced_data is None else int(traced_data.start_time)
            ),
            context=ctx,
        )
        _set_request_attributes(span, prepare_kwargs_for_shared_attributes(non_sentinel_kwargs), instance)
        span.set_attribute(ERROR_TYPE, e.__class__.__name__)
        span.record_exception(e)
        span.set_status(StatusCode.ERROR, str(e))
        if traced_data:
            set_data_attributes(traced_data, span)
        span.end()
        raise
    parsed_response = parse_response(response)

    existing_data = responses.get(parsed_response.id)
    if existing_data is None:
        existing_data = {}
    else:
        existing_data = existing_data.model_dump()

    request_tools = get_tools_from_kwargs(non_sentinel_kwargs)

    merged_tools = existing_data.get("tools", []) + request_tools

    try:
        parsed_response_output_text = None
        if hasattr(parsed_response, "output_text"):
            parsed_response_output_text = parsed_response.output_text
        else:
            try:
                parsed_response_output_text = parsed_response.output[0].content[0].text
            except Exception:
                pass
        traced_data = TracedData(
            start_time=existing_data.get("start_time", start_time),
            response_id=parsed_response.id,
            input=process_input(existing_data.get("input", non_sentinel_kwargs.get("input"))),
            instructions=existing_data.get("instructions", non_sentinel_kwargs.get("instructions")),
            tools=merged_tools if merged_tools else None,
            output_blocks={block.id: block for block in parsed_response.output}
            | existing_data.get("output_blocks", {}),
            usage=existing_data.get("usage", parsed_response.usage),
            output_text=existing_data.get("output_text", parsed_response_output_text),
            request_model=existing_data.get("request_model", non_sentinel_kwargs.get("model")),
            response_model=existing_data.get("response_model", parsed_response.model),
            # Reasoning attributes
            request_reasoning_summary=(
                non_sentinel_kwargs.get("reasoning", {}).get(
                    "summary", existing_data.get("request_reasoning_summary")
                )
            ),
            request_reasoning_effort=(
                non_sentinel_kwargs.get("reasoning", {}).get(
                    "effort", existing_data.get("request_reasoning_effort")
                )
            ),
            response_reasoning_effort=non_sentinel_kwargs.get("reasoning", {}).get("effort"),
            request_service_tier=existing_data.get("request_service_tier", non_sentinel_kwargs.get("service_tier")),
            response_service_tier=existing_data.get("response_service_tier", parsed_response.service_tier),
            # Capture trace context to maintain continuity across async operations
            trace_context=existing_data.get("trace_context", context_api.get_current()),
        )
        responses[parsed_response.id] = traced_data
    except Exception:
        return response

    if parsed_response.status == "completed":
        # Restore the original trace context to maintain trace continuity
        ctx = traced_data.trace_context if traced_data.trace_context else context_api.get_current()
        span = tracer.start_span(
            SPAN_NAME,
            kind=SpanKind.CLIENT,
            start_time=int(traced_data.start_time),
            context=ctx,
        )
        _set_request_attributes(span, prepare_kwargs_for_shared_attributes(non_sentinel_kwargs), instance)
        set_data_attributes(traced_data, span)
        span.end()

    return response


@dont_throw
@_with_tracer_wrapper
async def async_responses_get_or_create_wrapper(
    tracer: Tracer, wrapped, instance, args, kwargs
):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return await wrapped(*args, **kwargs)
    start_time = time.time_ns()

    # Remove OpenAI sentinel values (NOT_GIVEN, Omit) to allow chained .get() calls
    non_sentinel_kwargs = _sanitize_sentinel_values(kwargs)

    try:
        response = await wrapped(*args, **kwargs)
        if isinstance(response, (Stream, AsyncStream)):
            # Capture current trace context to maintain trace continuity
            ctx = context_api.get_current()
            span = tracer.start_span(
                SPAN_NAME,
                kind=SpanKind.CLIENT,
                start_time=start_time,
                context=ctx,
            )
            _set_request_attributes(span, prepare_kwargs_for_shared_attributes(non_sentinel_kwargs), instance)

            return ResponseStream(
                span=span,
                response=response,
                start_time=start_time,
                request_kwargs=non_sentinel_kwargs,
                tracer=tracer,
            )
    except Exception as e:
        response_id = non_sentinel_kwargs.get("response_id")
        existing_data = {}
        if response_id and response_id in responses:
            existing_data = responses[response_id].model_dump()
        try:
            traced_data = TracedData(
                start_time=existing_data.get("start_time", start_time),
                response_id=response_id or "",
                input=process_input(
                    non_sentinel_kwargs.get("input", existing_data.get("input", []))
                ),
                instructions=non_sentinel_kwargs.get(
                    "instructions", existing_data.get("instructions", "")
                ),
                tools=get_tools_from_kwargs(non_sentinel_kwargs) or existing_data.get("tools", []),
                output_blocks=existing_data.get("output_blocks", {}),
                usage=existing_data.get("usage"),
                output_text=non_sentinel_kwargs.get("output_text", existing_data.get("output_text")),
                request_model=non_sentinel_kwargs.get("model", existing_data.get("request_model")),
                response_model=existing_data.get("response_model"),
                # Reasoning attributes
                request_reasoning_summary=(
                    non_sentinel_kwargs.get("reasoning", {}).get(
                        "summary", existing_data.get("request_reasoning_summary")
                    )
                ),
                request_reasoning_effort=(
                    non_sentinel_kwargs.get("reasoning", {}).get(
                        "effort", existing_data.get("request_reasoning_effort")
                    )
                ),
                response_reasoning_effort=non_sentinel_kwargs.get("reasoning", {}).get("effort"),
                request_service_tier=non_sentinel_kwargs.get("service_tier"),
                response_service_tier=existing_data.get("response_service_tier"),
                # Capture trace context to maintain continuity
                trace_context=existing_data.get("trace_context", context_api.get_current()),
            )
        except Exception:
            traced_data = None

        # Restore the original trace context to maintain trace continuity
        ctx = (traced_data.trace_context if traced_data and traced_data.trace_context
               else context_api.get_current())
        span = tracer.start_span(
            SPAN_NAME,
            kind=SpanKind.CLIENT,
            start_time=(
                start_time if traced_data is None else int(traced_data.start_time)
            ),
            context=ctx,
        )
        _set_request_attributes(span, prepare_kwargs_for_shared_attributes(non_sentinel_kwargs), instance)
        span.set_attribute(ERROR_TYPE, e.__class__.__name__)
        span.record_exception(e)
        span.set_status(StatusCode.ERROR, str(e))
        if traced_data:
            set_data_attributes(traced_data, span)
        span.end()
        raise
    parsed_response = parse_response(response)

    existing_data = responses.get(parsed_response.id)
    if existing_data is None:
        existing_data = {}
    else:
        existing_data = existing_data.model_dump()

    request_tools = get_tools_from_kwargs(non_sentinel_kwargs)

    merged_tools = existing_data.get("tools", []) + request_tools

    try:
        parsed_response_output_text = None
        if hasattr(parsed_response, "output_text"):
            parsed_response_output_text = parsed_response.output_text
        else:
            try:
                parsed_response_output_text = parsed_response.output[0].content[0].text
            except Exception:
                pass

        traced_data = TracedData(
            start_time=existing_data.get("start_time", start_time),
            response_id=parsed_response.id,
            input=process_input(existing_data.get("input", non_sentinel_kwargs.get("input"))),
            instructions=existing_data.get("instructions", non_sentinel_kwargs.get("instructions")),
            tools=merged_tools if merged_tools else None,
            output_blocks={block.id: block for block in parsed_response.output}
            | existing_data.get("output_blocks", {}),
            usage=existing_data.get("usage", parsed_response.usage),
            output_text=existing_data.get("output_text", parsed_response_output_text),
            request_model=existing_data.get("request_model", non_sentinel_kwargs.get("model")),
            response_model=existing_data.get("response_model", parsed_response.model),
            # Reasoning attributes
            request_reasoning_summary=(
                non_sentinel_kwargs.get("reasoning", {}).get(
                    "summary", existing_data.get("request_reasoning_summary")
                )
            ),
            request_reasoning_effort=(
                non_sentinel_kwargs.get("reasoning", {}).get(
                    "effort", existing_data.get("request_reasoning_effort")
                )
            ),
            response_reasoning_effort=non_sentinel_kwargs.get("reasoning", {}).get("effort"),
            request_service_tier=existing_data.get("request_service_tier", non_sentinel_kwargs.get("service_tier")),
            response_service_tier=existing_data.get("response_service_tier", parsed_response.service_tier),
            # Capture trace context to maintain continuity across async operations
            trace_context=existing_data.get("trace_context", context_api.get_current()),
        )
        responses[parsed_response.id] = traced_data
    except Exception:
        return response

    if parsed_response.status == "completed":
        # Restore the original trace context to maintain trace continuity
        ctx = traced_data.trace_context if traced_data.trace_context else context_api.get_current()
        span = tracer.start_span(
            SPAN_NAME,
            kind=SpanKind.CLIENT,
            start_time=int(traced_data.start_time),
            context=ctx,
        )
        _set_request_attributes(span, prepare_kwargs_for_shared_attributes(non_sentinel_kwargs), instance)
        set_data_attributes(traced_data, span)
        span.end()

    return response


@dont_throw
@_with_tracer_wrapper
def responses_cancel_wrapper(tracer: Tracer, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    non_sentinel_kwargs = _sanitize_sentinel_values(kwargs)

    response = wrapped(*args, **kwargs)
    if isinstance(response, Stream):
        return response
    parsed_response = parse_response(response)
    existing_data = responses.pop(parsed_response.id, None)
    if existing_data is not None:
        # Restore the original trace context to maintain trace continuity
        ctx = existing_data.trace_context if existing_data.trace_context else context_api.get_current()
        span = tracer.start_span(
            SPAN_NAME,
            kind=SpanKind.CLIENT,
            start_time=existing_data.start_time,
            record_exception=True,
            context=ctx,
        )
        _set_request_attributes(span, prepare_kwargs_for_shared_attributes(non_sentinel_kwargs), instance)
        span.record_exception(Exception("Response cancelled"))
        set_data_attributes(existing_data, span)
        span.end()
    return response


@dont_throw
@_with_tracer_wrapper
async def async_responses_cancel_wrapper(
    tracer: Tracer, wrapped, instance, args, kwargs
):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return await wrapped(*args, **kwargs)

    non_sentinel_kwargs = _sanitize_sentinel_values(kwargs)

    response = await wrapped(*args, **kwargs)
    if isinstance(response, (Stream, AsyncStream)):
        return response
    parsed_response = parse_response(response)
    existing_data = responses.pop(parsed_response.id, None)
    if existing_data is not None:
        # Restore the original trace context to maintain trace continuity
        ctx = existing_data.trace_context if existing_data.trace_context else context_api.get_current()
        span = tracer.start_span(
            SPAN_NAME,
            kind=SpanKind.CLIENT,
            start_time=existing_data.start_time,
            record_exception=True,
            context=ctx,
        )
        _set_request_attributes(span, prepare_kwargs_for_shared_attributes(non_sentinel_kwargs), instance)
        span.record_exception(Exception("Response cancelled"))
        set_data_attributes(existing_data, span)
        span.end()
    return response


class ResponseStream(ObjectProxy):
    """Proxy class for streaming responses to capture telemetry data"""

    _span = None
    _start_time = None
    _request_kwargs = None
    _tracer = None
    _traced_data = None

    def __init__(
        self,
        span,
        response,
        start_time=None,
        request_kwargs=None,
        tracer=None,
        traced_data=None,
    ):
        super().__init__(response)
        self._span = span
        self._start_time = start_time
        # Filter sentinel values (defensive, in case called directly without prior filtering)
        self._request_kwargs = _sanitize_sentinel_values(request_kwargs or {})
        self._tracer = tracer
        self._traced_data = traced_data or TracedData(
            start_time=start_time,
            response_id="",
            input=process_input(self._request_kwargs.get("input", [])),
            instructions=self._request_kwargs.get("instructions"),
            tools=get_tools_from_kwargs(self._request_kwargs),
            output_blocks={},
            usage=None,
            output_text="",
            request_model=self._request_kwargs.get("model", ""),
            response_model="",
            request_reasoning_summary=self._request_kwargs.get("reasoning", {}).get(
                "summary"
            ),
            request_reasoning_effort=self._request_kwargs.get("reasoning", {}).get(
                "effort"
            ),
            response_reasoning_effort=None,
            request_service_tier=self._request_kwargs.get("service_tier"),
            response_service_tier=None,
        )

        self._complete_response_data = None
        self._output_text = ""

        self._cleanup_completed = False
        self._cleanup_lock = threading.Lock()

    def __del__(self):
        """Cleanup when object is garbage collected"""
        if hasattr(self, "_cleanup_completed") and not self._cleanup_completed:
            self._ensure_cleanup()

    def __enter__(self):
        """Context manager entry"""
        if hasattr(self.__wrapped__, "__enter__"):
            self.__wrapped__.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit"""
        suppress = False
        try:
            if exc_type is not None:
                self._handle_exception(exc_val)
            else:
                self._process_complete_response()
        finally:
            if hasattr(self.__wrapped__, "__exit__"):
                suppress = bool(self.__wrapped__.__exit__(exc_type, exc_val, exc_tb))
        return suppress

    async def __aenter__(self):
        """Async context manager entry"""
        if hasattr(self.__wrapped__, "__aenter__"):
            await self.__wrapped__.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit"""
        suppress = False
        try:
            if exc_type is not None:
                self._handle_exception(exc_val)
            else:
                self._process_complete_response()
        finally:
            if hasattr(self.__wrapped__, "__aexit__"):
                suppress = bool(await self.__wrapped__.__aexit__(exc_type, exc_val, exc_tb))
        return suppress

    def close(self):
        try:
            self._ensure_cleanup()
        finally:
            if hasattr(self.__wrapped__, "close"):
                return self.__wrapped__.close()

    async def aclose(self):
        try:
            self._ensure_cleanup()
        finally:
            if hasattr(self.__wrapped__, "aclose"):
                return await self.__wrapped__.aclose()

    def __iter__(self):
        """Synchronous iterator"""
        return self

    def __next__(self):
        """Synchronous iteration"""
        try:
            chunk = self.__wrapped__.__next__()
        except StopIteration:
            self._process_complete_response()
            raise
        except Exception as e:
            self._handle_exception(e)
            raise
        else:
            self._process_chunk(chunk)
            return chunk

    def __aiter__(self):
        """Async iterator"""
        return self

    async def __anext__(self):
        """Async iteration"""
        try:
            chunk = await self.__wrapped__.__anext__()
        except StopAsyncIteration:
            self._process_complete_response()
            raise
        except Exception as e:
            self._handle_exception(e)
            raise
        else:
            self._process_chunk(chunk)
            return chunk

    def _process_chunk(self, chunk):
        """Process a streaming chunk"""
        if hasattr(chunk, "type"):
            if chunk.type == "response.output_text.delta":
                if hasattr(chunk, "delta") and chunk.delta:
                    self._output_text += chunk.delta
            elif chunk.type == "response.completed" and hasattr(chunk, "response"):
                self._complete_response_data = chunk.response

        if hasattr(chunk, "delta"):
            if hasattr(chunk.delta, "text") and chunk.delta.text:
                self._output_text += chunk.delta.text

        if hasattr(chunk, "response") and chunk.response:
            self._complete_response_data = chunk.response

    @dont_throw
    def _process_complete_response(self):
        """Process the complete response and emit span"""
        with self._cleanup_lock:
            if self._cleanup_completed:
                return

            try:
                if self._complete_response_data:
                    parsed_response = parse_response(self._complete_response_data)

                    self._traced_data.response_id = parsed_response.id
                    self._traced_data.response_model = parsed_response.model
                    self._traced_data.output_text = self._output_text

                    if parsed_response.usage:
                        self._traced_data.usage = parsed_response.usage

                    if parsed_response.output:
                        self._traced_data.output_blocks = {
                            block.id: block for block in parsed_response.output
                        }

                    responses[parsed_response.id] = self._traced_data

                set_data_attributes(self._traced_data, self._span)
                self._span.set_status(StatusCode.OK)
                self._span.end()
                self._cleanup_completed = True

            except Exception as e:
                if self._span and self._span.is_recording():
                    self._span.set_attribute(ERROR_TYPE, e.__class__.__name__)
                    self._span.set_status(StatusCode.ERROR, str(e))
                    self._span.end()
                self._cleanup_completed = True

    @dont_throw
    def _handle_exception(self, exception):
        """Handle exceptions during streaming"""
        with self._cleanup_lock:
            if self._cleanup_completed:
                return

            if self._span and self._span.is_recording():
                self._span.set_attribute(ERROR_TYPE, exception.__class__.__name__)
                self._span.record_exception(exception)
                self._span.set_status(StatusCode.ERROR, str(exception))
                self._span.end()

            self._cleanup_completed = True

    @dont_throw
    def _ensure_cleanup(self):
        """Ensure cleanup happens even if stream is not fully consumed"""
        with self._cleanup_lock:
            if self._cleanup_completed:
                return

            try:
                if self._span and self._span.is_recording():
                    set_data_attributes(self._traced_data, self._span)
                    self._span.set_status(StatusCode.OK)
                    self._span.end()

                self._cleanup_completed = True

            except Exception:
                self._cleanup_completed = True



import os
import logging
from typing import Optional
from overmind import layers
from overmind.overmind_sdk import init, get_tracer

logger = logging.getLogger(__name__)


class OvermindClient:
    def __init__(
        self,
        overmind_api_key: Optional[str] = None,
        overmind_base_url: Optional[str] = None,
    ):
        self.overmind_api_key = overmind_api_key or os.getenv(
            "OVERMIND_API_KEY",
        )
        if not self.overmind_api_key:
            raise ValueError("OVERMIND_API_KEY is not set")

        self.overmind_base_url = overmind_base_url or os.getenv(
            "OVERMIND_API_URL",
        )

        init(overmind_api_key, traces_base_url=self.overmind_base_url)

        logger.debug("OVERMIND_API_URL is set to: %s", self.overmind_base_url)

    def run_layer(self, *, policy, input_data, **kwargs):
        layer = get_layer(policy, **kwargs)
        with get_tracer().start_as_current_span(f"execute_policy_{policy}") as span:
            response = layer.run(input_data=input_data, **kwargs)
            span.set_attribute("inputs", str(input_data))
            span.set_attribute("outputs", str(response.processed_data))
            span.set_attribute("policy_outcome", response.overall_policy_outcome)
            span.set_attribute("policy_results", str(response.policy_results))

        return response


def get_layer(name: str, **kwargs) -> layers.GenericOvermindLayer:
    # layer factory
    if name == "anonymize_pii":
        return layers.AnonymizePIILayer(**kwargs)
    if name == "reject_irrelevant_answer":
        return layers.RejectIrrelevantAnswersLayer(**kwargs)
    if name == "reject_prompt_injection":
        return layers.RejectPromptInjectionLayer(**kwargs)
    if name == "reject_llm_judge_with_criteria":
        return layers.LLMJudgeScorerLayer(**kwargs)
    else:
        raise ValueError(f"Invalid layer name: {name}")



import logging
import os
from typing import Optional
from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import Resource

from overmind.utils.api_settings import get_api_settings

logger = logging.getLogger(__name__)

# Global state to track initialization
_initialized = False
_tracer: Optional[trace.Tracer] = None


def init(
    overmind_api_key: Optional[str] = None,
    traces_base_url: Optional[str] = None,
    service_name: Optional[str] = None,
    environment: Optional[str] = None,
    processes_sample_rate: float = 1.0,
    capture_request_body: bool = True,
    capture_response_body: bool = True,
) -> None:
    """
    Initialize the Overmind SDK for automatic monitoring.
    
    Call this once at application startup, before creating your FastAPI app.
    
    Example:
        import overmind
        overmind.init(service_name="my-backend")
        
        from fastapi import FastAPI
        app = FastAPI()

    Args:
        overmind_api_key: Your Overmind API key. If not provided, uses OVERMIND_API_KEY env var.
        traces_base_url: Base URL for traces. If not provided, uses OVERMIND_TRACES_URL env var.
        service_name: Name of your service (appears in traces). Defaults to OVERMIND_SERVICE_NAME 
                      env var or "unknown-service".
        environment: Environment name (e.g., "production", "staging"). Defaults to 
                     OVERMIND_ENVIRONMENT env var or "development".
        processes_sample_rate: Sampling rate for traces (0.0 to 1.0). Default 1.0 captures all.
        capture_request_body: Whether to capture OpenAI request payloads. Default True.
        capture_response_body: Whether to capture OpenAI response payloads. Default True.
    """
    global _initialized, _tracer
    
    if _initialized:
        logger.debug("Overmind SDK already initialized, skipping.")
        return
    
    try:
        api_key, _, traces_url = get_api_settings(
            overmind_api_key=overmind_api_key,
            traces_base_url=traces_base_url,
        )
    except Exception as e:
        logger.error(f"Failed to initialize Overmind SDK settings: {e}")
        return

    # Resolve service name and environment
    resolved_service_name = (
        service_name 
        or os.environ.get("OVERMIND_SERVICE_NAME") 
        or os.environ.get("SERVICE_NAME")
        or "unknown-service"
    )
    resolved_environment = (
        environment 
        or os.environ.get("OVERMIND_ENVIRONMENT") 
        or os.environ.get("ENVIRONMENT")
        or "development"
    )

    endpoint = f"{traces_url}/api/v1/traces/create"
    
    # Configure OpenTelemetry Provider with rich resource attributes
    resource = Resource.create({
        "service.name": resolved_service_name,
        "service.version": os.environ.get("SERVICE_VERSION", "unknown"),
        "deployment.environment": resolved_environment,
        "overmind.sdk.name": "overmind-python",
        "overmind.sdk.version": "0.1.15",
    })
    
    provider = TracerProvider(resource=resource)
    
    # Configure OTLP Exporter
    headers = {"X-API-Token": api_key}
    
    otlp_exporter = OTLPSpanExporter(endpoint=endpoint, headers=headers)
    span_processor = BatchSpanProcessor(otlp_exporter)
    provider.add_span_processor(span_processor)
    
    # Set global Trace Provider
    trace.set_tracer_provider(provider)
    
    # Store tracer for custom spans
    _tracer = trace.get_tracer("overmind", "0.1.15")

    # Instrument FastAPI (for HTTP request tracing)
    try:
        from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
        FastAPIInstrumentor().instrument()
        logger.info("Overmind SDK: FastAPI instrumentation enabled.")
    except ImportError:
        logger.debug("Overmind SDK: opentelemetry-instrumentation-fastapi not found, skipping.")
    except Exception as e:
        logger.warning(f"Overmind SDK: Failed to instrument FastAPI: {e}")

    # Instrument OpenAI (for LLM call tracing) - using vendored patched version
    try:
        from overmind._vendor.opentelemetry_instrumentation_openai import OpenAIInstrumentor
        OpenAIInstrumentor().instrument()
        logger.info("Overmind SDK: OpenAI instrumentation enabled (patched).")
    except ImportError:
        logger.debug("Overmind SDK: opentelemetry-instrumentation-openai not found, skipping.")
    except Exception as e:
        logger.warning(f"Overmind SDK: Failed to instrument OpenAI: {e}")

    # Instrument requests library (for outgoing HTTP calls)
    try:
        from opentelemetry.instrumentation.requests import RequestsInstrumentor
        RequestsInstrumentor().instrument()
        logger.info("Overmind SDK: Requests instrumentation enabled.")
    except ImportError:
        logger.debug("Overmind SDK: opentelemetry-instrumentation-requests not found, skipping.")
    except Exception as e:
        logger.warning(f"Overmind SDK: Failed to instrument Requests: {e}")

    # Instrument httpx (async HTTP client, used by newer OpenAI SDK)
    try:
        from opentelemetry.instrumentation.httpx import HTTPXClientInstrumentor
        HTTPXClientInstrumentor().instrument()
        logger.info("Overmind SDK: HTTPX instrumentation enabled.")
    except ImportError:
        logger.debug("Overmind SDK: opentelemetry-instrumentation-httpx not found, skipping.")
    except Exception as e:
        logger.warning(f"Overmind SDK: Failed to instrument HTTPX: {e}")

    # Instrument logging (to correlate logs with traces)
    try:
        from opentelemetry.instrumentation.logging import LoggingInstrumentor
        LoggingInstrumentor().instrument(set_logging_format=True)
        logger.info("Overmind SDK: Logging instrumentation enabled.")
    except ImportError:
        logger.debug("Overmind SDK: opentelemetry-instrumentation-logging not found, skipping.")
    except Exception as e:
        logger.warning(f"Overmind SDK: Failed to instrument Logging: {e}")

    _initialized = True
    logger.info(
        f"Overmind SDK initialized: service={resolved_service_name}, "
        f"environment={resolved_environment}"
    )


def get_tracer() -> trace.Tracer:
    """
    Get the Overmind tracer for creating custom spans.
    
    Example:
        tracer = overmind.get_tracer()
        with tracer.start_as_current_span("my-operation") as span:
            span.set_attribute("user.id", user_id)
            # ... your code ...
    
    Returns:
        OpenTelemetry Tracer instance.
        
    Raises:
        RuntimeError: If SDK not initialized.
    """
    if not _initialized or _tracer is None:
        raise RuntimeError(
            "Overmind SDK not initialized. Call overmind.init() first."
        )
    return _tracer


def set_user(user_id: str, email: Optional[str] = None, username: Optional[str] = None) -> None:
    """
    Associate current trace with a user (like Sentry's set_user).
    
    Call this in your request handler to tag traces with user info.
    
    Example:
        @app.middleware("http")
        async def add_user_context(request: Request, call_next):
            if request.state.user:
                overmind.set_user(user_id=request.state.user.id)
            return await call_next(request)

    Args:
        user_id: Unique user identifier.
        email: Optional user email.
        username: Optional username.
    """
    span = trace.get_current_span()
    if span.is_recording():
        span.set_attribute("user.id", user_id)
        if email:
            span.set_attribute("user.email", email)
        if username:
            span.set_attribute("user.username", username)


def set_tag(key: str, value: str) -> None:
    """
    Add a custom tag to the current span.
    
    Example:
        overmind.set_tag("feature.flag", "new-checkout-flow")
        overmind.set_tag("tenant.id", tenant_id)

    Args:
        key: Tag name.
        value: Tag value.
    """
    span = trace.get_current_span()
    if span.is_recording():
        span.set_attribute(key, value)


def capture_exception(exception: Exception) -> None:
    """
    Record an exception on the current span.
    
    Example:
        try:
            risky_operation()
        except Exception as e:
            overmind.capture_exception(e)
            raise

    Args:
        exception: The exception to record.
    """
    span = trace.get_current_span()
    if span.is_recording():
        span.record_exception(exception)
        span.set_status(trace.Status(trace.StatusCode.ERROR, str(exception)))



"""
Overmind Python Client

A Python client for the Overmind API that provides easy access to AI provider endpoints
with policy enforcement and automatic observability.
"""

from .client import OvermindClient
from .exceptions import OvermindAPIError, OvermindAuthenticationError, OvermindError
from .overmind_sdk import init, get_tracer, set_user, set_tag, capture_exception

__version__ = "0.1.0"
__all__ = [
    "OvermindClient",
    "OvermindError",
    "OvermindAPIError",
    "OvermindAuthenticationError",
    "init",
    "get_tracer",
    "set_user",
    "set_tag",
    "capture_exception",
]



from collections.abc import Iterator
import json
from pathlib import Path
from overmind.overmind_sdk import init, get_tracer
from tqdm import tqdm
from opentelemetry import trace
from typing import Dict, Optional
from pydantic import BaseModel, ConfigDict, Field
import csv
from opentelemetry.trace import SpanKind


def get_log_item_model(mapping: Dict[str, str] = None):
    if mapping is None:
        mapping = {}

    def get_field(name: str):
        return mapping.get(name, name)

    class LogItem(BaseModel):
        input: object = Field(..., alias=get_field("input"))
        output: object = Field(..., alias=get_field("output"))

        start_time: int = Field(..., alias=get_field("start_time"))
        end_time: int = Field(..., alias=get_field("end_time"))

        # custom IDs (must be valid 32/16 hex chars)
        trace_state: Optional[str] = Field(
            default=None, alias=get_field("trace_state"), max_length=32, min_length=16
        )
        trace_id: Optional[str] = Field(
            default=None, alias=get_field("trace_id"), max_length=32, min_length=16
        )
        span_id: Optional[str] = Field(
            default=None, alias=get_field("span_id"), max_length=32, min_length=16
        )
        parent_span_id: Optional[str] = Field(
            default=None,
            alias=get_field("parent_span_id"),
            max_length=32,
            min_length=16,
        )

        name: Optional[str] = Field(
            default="log-ingestion-service", alias=get_field("name")
        )
        kind: int = Field(default=2, alias=get_field("kind"))

        status_code: int = Field(default=0, alias=get_field("status_code"))
        status_message: str = Field(default="", alias=get_field("status_message"))
        extra_attributes: Dict[str, str] = Field(
            ..., default_factory=dict, alias=get_field("extra_attributes")
        )

        model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    return LogItem


def process_log_item(item: dict, mapping: Dict[str, str]):
    log_model = get_log_item_model(mapping)
    log_item = log_model.model_validate(item, by_alias=True)

    ctx = {}
    if log_item.trace_id:
        ctx["trace_id"] = log_item.trace_id
    if log_item.span_id:
        ctx["span_id"] = log_item.span_id
    if log_item.trace_state:
        ctx["trace_state"] = log_item.trace_state

    context = None
    if ctx:
        context = trace.set_span_in_context(
            trace.NonRecordingSpan(
                trace.SpanContext(
                    **ctx,
                    is_remote=True,
                )
            )
        )

    tracer = get_tracer()
    span = tracer.start_span(
        name=log_item.name or "log-ingestion-service",
        start_time=log_item.start_time,
        kind=SpanKind(log_item.kind),
        attributes=log_item.extra_attributes,
        context=context,
    )
    span.set_status(log_item.status_code, log_item.status_message)
    span.set_attribute("inputs", str(log_item.input))
    span.set_attribute("outputs", str(log_item.output))
    span.end(end_time=log_item.end_time)


def load_from_jsonl(filepath: str) -> Iterator[dict]:
    with open(filepath, "r") as f:
        for line in f:
            yield json.loads(line)


def load_from_json(filepath: str) -> Iterator[dict]:
    with open(filepath, "r") as f:
        data = json.load(f)
        for item in data:
            yield item


def load_from_csv(filepath: str) -> Iterator[dict]:
    with open(filepath, "r", newline="") as f:
        reader = csv.DictReader(f)
        for row in reader:
            yield row


def ingest_logs(filepath: str, mapping: Dict[str, str], **kwargs):
    if kwargs.get("overmind_api_key"):
        init(overmind_api_key=kwargs.get("overmind_api_key"))
    else:
        init()

    suffix = Path(filepath).suffix
    items: list[dict] = []
    if suffix == ".jsonl":
        items = load_from_jsonl(filepath)
    elif suffix == ".json":
        items = load_from_json(filepath)
    elif suffix == ".csv":
        items = load_from_csv(filepath)
    else:
        raise ValueError(f"Unsupported file extension: {suffix}")

    for i, dict_item in tqdm(enumerate(items)):
        try:
            process_log_item(dict_item, mapping)
        except Exception as trace_ex:
            print(f"Failed to record span for line {i}: {trace_ex}")
            continue


if __name__ == "__main__":
    data = "tests/data/logs_mapped.csv"  # Fixed path if running from project root

    mapping = {
        "input": "llm_input",
        "output": "llm_output",
        "start_time": "start_time_nano",
        "end_time": "end_time_nano",
    }
    ingest_logs(data, mapping)



from rich.console import Console, Group
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from rich.syntax import Syntax
import json


res_map = {
    "passed": {"color": "green", "emoji": "✅"},
    "altered": {"color": "yellow", "emoji": "⚠️"},
    "rejected": {"color": "red", "emoji": "❌"},
}


def _get_outcome_str(outcome_str: str) -> str:
    style = res_map.get(outcome_str, {}).get("color", "white")
    emoji = res_map.get(outcome_str, {}).get("emoji", "➡️")
    outcome_text = Text(f"{emoji} {outcome_str.upper()}", style=style)
    return outcome_text


def summarize_proxy_run(result) -> None:
    # -- 2. Set up Rich console and formatting maps --
    console = Console()

    # -- 3. Build the Rich renderable components --

    # A Table for Invocation Results
    inv_table = Table(show_header=True, header_style="bold white", title="Results")
    inv_table.add_column("Component", style="dim")
    inv_table.add_column("Outcome", justify="center")

    overall_input_layer_result = result.input_layer_results.get(
        "overall_policy_outcome"
    )
    if overall_input_layer_result:
        inv_table.add_row("Input Layer", _get_outcome_str(overall_input_layer_result))

    overall_output_layer_result = result.output_layer_results.get(
        "overall_policy_outcome"
    )
    if overall_output_layer_result:
        inv_table.add_row("Output Layer", _get_outcome_str(overall_output_layer_result))

    # A Table for Policy Results
    pol_table = Table(
        show_header=True, header_style="bold white", title="\nPolicy Results"
    )
    pol_table.add_column("Layer", style="dim", width=10)
    pol_table.add_column("Policy", style="dim", width=20)
    pol_table.add_column("Details", width=70)

    for policy, outcome in result.input_layer_results.get("policy_results", {}).items():
        # Assuming the outcome is a dict, format it nicely with Syntax
        outcome_str = json.dumps(outcome, indent=2)
        outcome_syntax = Syntax(
            outcome_str, "json", theme="solarized-dark", word_wrap=True
        )
        pol_table.add_row("In", policy, outcome_syntax)

    for policy, outcome in result.output_layer_results.get(
        "policy_results", {}
    ).items():
        # Assuming the outcome is a dict, format it nicely with Syntax
        outcome_str = json.dumps(outcome, indent=2)
        outcome_syntax = Syntax(
            outcome_str, "json", theme="solarized-dark", word_wrap=True
        )
        pol_table.add_row("Out", policy, outcome_syntax)

    # Syntax Panels for Processed Input and Output
    input_color = res_map.get(overall_input_layer_result, {}).get("color", "white")
    output_color = res_map.get(overall_output_layer_result, {}).get("color", "white")

    input_panel = Panel(
        Syntax(
            json.dumps(result.processed_input, indent=2)[:2000],
            "plain",
            theme="solarized-dark",
            line_numbers=False,
            word_wrap=True,
        ),
        title="[bold]Processed Input[/bold]",
        border_style=input_color,  # The border color reflects the outcome!
        title_align="left",
    )

    output_panel = Panel(
        Syntax(
            json.dumps(result.processed_output)[:2000],
            "plain",
            theme="solarized-dark",
            line_numbers=False,
            word_wrap=True,
        ),
        title="[bold]Processed Output[/bold]",
        border_style=output_color,  # The border color reflects the outcome!
        title_align="left",
    )

    # -- 4. Group components and print inside a final Panel --
    final_group = Group(
        inv_table,
        pol_table,
        input_panel,
        output_panel,
    )

    console.print(
        Panel(
            final_group,
            title="[bold cyan]Proxy Run Summary[/bold cyan]",
            border_style="green",
        )
    )



from .openai import OpenAI, AsyncOpenAI


__all__ = ["OpenAI", "AsyncOpenAI"]

try:
    from .anthropic import Anthropic

    __all__.append("Anthropic")
except ImportError:
    pass

try:
    from .google import Client as GoogleClient

    __all__.append("GoogleClient")
except ImportError:
    pass



"""
Instrumentation wrappers for OpenAI Realtime API.

The Realtime API uses WebSocket connections for low-latency, multi-modal conversations.
This module provides wrappers to instrument:
- Session lifecycle (connect/disconnect)
- Response cycles (response.create -> response.done)
- Tool/function calls

Key concepts:
- A "session" span covers the entire WebSocket connection lifecycle
- A "response" span covers each response.create -> response.done cycle
- Token usage is captured from response.done events
"""

import json
import time
from typing import Optional

from opentelemetry import context as context_api
from opentelemetry import trace
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY
from opentelemetry.semconv._incubating.attributes import (
    gen_ai_attributes as GenAIAttributes,
)
from opentelemetry.semconv_ai import SpanAttributes
from opentelemetry.trace import SpanKind, Status, StatusCode, Tracer

from ..shared import (
    _set_span_attribute,
    model_as_dict,
)
from ..utils import (
    _with_tracer_wrapper,
    dont_throw,
    should_send_prompts,
)


SPAN_NAME_SESSION = "openai.session"
SPAN_NAME_RESPONSE = "openai.realtime"


class RealtimeSessionState:
    """Tracks state for a realtime session to correlate events with spans."""

    def __init__(self, tracer: Tracer, model: str):
        self.tracer = tracer
        self.model = model
        self.session_span = None
        self.response_span = None
        self.response_start_time = None
        self.session_config: dict = {}
        self.current_response_id: Optional[str] = None
        self.accumulated_text: str = ""
        self.accumulated_audio_transcript: str = ""
        self.function_calls: list = []
        self.input_text: str = ""
        self.trace_context = None


class RealtimeEventProcessor:
    """Shared event processing logic for realtime connections."""

    def __init__(self, state: RealtimeSessionState):
        self._state = state

    @dont_throw
    def process_event(self, event):
        """Process incoming server events."""
        event_type = getattr(event, "type", None)
        if not event_type:
            return

        if event_type == "session.created":
            self.handle_session_created(event)
        elif event_type == "session.updated":
            self.handle_session_updated(event)
        elif event_type == "response.created":
            self.handle_response_created(event)
        elif event_type == "response.text.delta":
            self.handle_text_delta(event)
        elif event_type == "response.audio_transcript.delta":
            self.handle_audio_transcript_delta(event)
        elif event_type == "response.function_call_arguments.done":
            self.handle_function_call_done(event)
        elif event_type == "response.done":
            self.handle_response_done(event)
        elif event_type == "error":
            self.handle_error_event(event)

    @dont_throw
    def handle_session_created(self, event):
        """Handle session.created event."""
        if hasattr(event, "session"):
            session = event.session
            if hasattr(session, "model"):
                self._state.model = session.model
            if hasattr(session, "modalities"):
                self._state.session_config["modalities"] = session.modalities
            if hasattr(session, "instructions"):
                self._state.session_config["instructions"] = session.instructions

        if self._state.session_span and self._state.session_span.is_recording():
            _set_span_attribute(
                self._state.session_span,
                GenAIAttributes.GEN_AI_REQUEST_MODEL,
                self._state.model,
            )

    @dont_throw
    def handle_session_updated(self, event):
        """Handle session.updated event."""
        if hasattr(event, "session"):
            session = event.session
            session_dict = (
                model_as_dict(session) if hasattr(session, "__dict__") else {}
            )
            self._state.session_config.update(session_dict)

            if self._state.session_span and self._state.session_span.is_recording():
                if hasattr(session, "modalities"):
                    _set_span_attribute(
                        self._state.session_span,
                        f"{SpanAttributes.LLM_REQUEST_TYPE}.modalities",
                        json.dumps(session.modalities) if session.modalities else None,
                    )
                if hasattr(session, "temperature") and session.temperature is not None:
                    _set_span_attribute(
                        self._state.session_span,
                        GenAIAttributes.GEN_AI_REQUEST_TEMPERATURE,
                        session.temperature,
                    )

    @dont_throw
    def handle_response_created(self, event, start_span_if_none: bool = False):
        """Handle response.created event - response cycle started."""
        if hasattr(event, "response") and hasattr(event.response, "id"):
            self._state.current_response_id = event.response.id

        # Optionally start span if not already started (used by iterator)
        if start_span_if_none and self._state.response_span is None:
            self.start_response_span()

    @dont_throw
    def handle_text_delta(self, event):
        """Handle response.text.delta event - accumulate text."""
        if hasattr(event, "delta"):
            self._state.accumulated_text += event.delta

    @dont_throw
    def handle_audio_transcript_delta(self, event):
        """Handle response.audio_transcript.delta event."""
        if hasattr(event, "delta"):
            self._state.accumulated_audio_transcript += event.delta

    @dont_throw
    def handle_function_call_done(self, event):
        """Handle response.function_call_arguments.done event."""
        self._state.function_calls.append({
            "name": getattr(event, "name", None),
            "call_id": getattr(event, "call_id", None),
            "arguments": getattr(event, "arguments", None),
        })

    @dont_throw
    def handle_response_done(self, event):
        """Handle response.done event - end the response span."""
        if self._state.response_span is None:
            return

        span = self._state.response_span

        if span.is_recording():
            # Set response attributes
            if hasattr(event, "response"):
                response = event.response
                _set_span_attribute(
                    span,
                    GenAIAttributes.GEN_AI_RESPONSE_ID,
                    getattr(response, "id", None),
                )

                # Set usage metrics
                if hasattr(response, "usage") and response.usage:
                    usage = response.usage
                    _set_span_attribute(
                        span,
                        GenAIAttributes.GEN_AI_USAGE_INPUT_TOKENS,
                        getattr(usage, "input_tokens", None),
                    )
                    _set_span_attribute(
                        span,
                        GenAIAttributes.GEN_AI_USAGE_OUTPUT_TOKENS,
                        getattr(usage, "output_tokens", None),
                    )
                    total = (getattr(usage, "input_tokens", 0) or 0) + (
                        getattr(usage, "output_tokens", 0) or 0
                    )
                    _set_span_attribute(
                        span,
                        SpanAttributes.LLM_USAGE_TOTAL_TOKENS,
                        total,
                    )

                # Set output content if tracing is enabled
                if should_send_prompts():
                    # Always set role for completions
                    _set_span_attribute(
                        span,
                        f"{GenAIAttributes.GEN_AI_COMPLETION}.0.role",
                        "assistant",
                    )

                    # Set content (text or audio transcript)
                    if self._state.accumulated_text:
                        _set_span_attribute(
                            span,
                            f"{GenAIAttributes.GEN_AI_COMPLETION}.0.content",
                            self._state.accumulated_text,
                        )
                    elif self._state.accumulated_audio_transcript:
                        _set_span_attribute(
                            span,
                            f"{GenAIAttributes.GEN_AI_COMPLETION}.0.content",
                            self._state.accumulated_audio_transcript,
                        )

                    # Set tool calls and finish_reason
                    if self._state.function_calls:
                        _set_span_attribute(
                            span,
                            f"{GenAIAttributes.GEN_AI_COMPLETION}.0.finish_reason",
                            "tool_calls",
                        )
                        for i, call in enumerate(self._state.function_calls):
                            _set_span_attribute(
                                span,
                                f"{GenAIAttributes.GEN_AI_COMPLETION}.0.tool_calls.{i}.type",
                                "function",
                            )
                            _set_span_attribute(
                                span,
                                f"{GenAIAttributes.GEN_AI_COMPLETION}.0.tool_calls.{i}.name",
                                call.get("name"),
                            )
                            _set_span_attribute(
                                span,
                                f"{GenAIAttributes.GEN_AI_COMPLETION}.0.tool_calls.{i}.id",
                                call.get("call_id"),
                            )
                            _set_span_attribute(
                                span,
                                f"{GenAIAttributes.GEN_AI_COMPLETION}.0.tool_calls.{i}.arguments",
                                call.get("arguments"),
                            )
                    else:
                        _set_span_attribute(
                            span,
                            f"{GenAIAttributes.GEN_AI_COMPLETION}.0.finish_reason",
                            "stop",
                        )

            span.set_status(Status(StatusCode.OK))

        span.end()
        self.reset_response_state()

    @dont_throw
    def handle_error_event(self, event):
        """Handle error event."""
        error_msg = str(getattr(event, "error", "Unknown error"))

        if self._state.response_span and self._state.response_span.is_recording():
            self._state.response_span.set_status(Status(StatusCode.ERROR, error_msg))
            self._state.response_span.record_exception(Exception(error_msg))
            self._state.response_span.end()
            self.reset_response_state()

    def start_response_span(self, end_existing: bool = False, set_input: bool = False):
        """Start a new response span."""
        if self._state.response_span is not None:
            if end_existing:
                # End any existing response span with OK status
                self._state.response_span.set_status(Status(StatusCode.OK))
                self._state.response_span.end()
                self.reset_response_state()
            else:
                return  # Don't start a new span if one exists and end_existing=False

        self._state.response_start_time = time.time_ns()

        ctx = self._state.trace_context or context_api.get_current()

        self._state.response_span = self._state.tracer.start_span(
            SPAN_NAME_RESPONSE,
            kind=SpanKind.CLIENT,
            start_time=self._state.response_start_time,
            context=ctx,
        )

        if self._state.response_span.is_recording():
            _set_span_attribute(
                self._state.response_span,
                GenAIAttributes.GEN_AI_SYSTEM,
                "openai",
            )
            _set_span_attribute(
                self._state.response_span,
                GenAIAttributes.GEN_AI_REQUEST_MODEL,
                self._state.model,
            )
            _set_span_attribute(
                self._state.response_span,
                SpanAttributes.LLM_REQUEST_TYPE,
                "realtime",
            )

            # Set input if available and requested
            if set_input and should_send_prompts() and self._state.input_text:
                _set_span_attribute(
                    self._state.response_span,
                    f"{GenAIAttributes.GEN_AI_PROMPT}.0.content",
                    self._state.input_text,
                )
                _set_span_attribute(
                    self._state.response_span,
                    f"{GenAIAttributes.GEN_AI_PROMPT}.0.role",
                    "user",
                )

    def reset_response_state(self):
        """Reset state for the next response cycle."""
        self._state.response_span = None
        self._state.response_start_time = None
        self._state.current_response_id = None
        self._state.accumulated_text = ""
        self._state.accumulated_audio_transcript = ""
        self._state.function_calls = []
        self._state.input_text = ""


class RealtimeConnectionWrapper:
    """
    Wrapper for OpenAI Realtime connection that instruments events.

    Wraps the connection object returned by client.beta.realtime.connect()
    to capture telemetry for session lifecycle and response cycles.
    """

    def __init__(self, connection, state: RealtimeSessionState):
        self._connection = connection
        self._state = state
        self._processor = RealtimeEventProcessor(state)
        self._closed = False

    def __getattr__(self, name):
        """Delegate attribute access to the wrapped connection."""
        return getattr(self._connection, name)

    @property
    def session(self):
        """Return a wrapped session object."""
        return RealtimeSessionWrapper(self._connection.session, self._state)

    @property
    def conversation(self):
        """Return a wrapped conversation object."""
        return RealtimeConversationWrapper(self._connection.conversation, self._state)

    @property
    def response(self):
        """Return a wrapped response object."""
        return RealtimeResponseWrapper(
            self._connection.response, self._state, self._processor
        )

    async def recv(self):
        """Receive and process an event."""
        event = await self._connection.recv()
        self._processor.process_event(event)
        return event

    def recv_bytes(self):
        """Delegate to wrapped connection."""
        return self._connection.recv_bytes()

    async def send(self, event):
        """Send an event, tracking response.create."""
        self._process_outgoing_event(event)
        return await self._connection.send(event)

    def __aiter__(self):
        """Return async iterator for events."""
        return RealtimeEventIterator(self._connection, self._state, self._processor)

    async def close(self):
        """Close the connection and end the session span."""
        if not self._closed:
            self._closed = True
            self._end_session_span()
        return await self._connection.close()

    async def __aenter__(self):
        """Async context manager entry."""
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit."""
        if not self._closed:
            self._closed = True
            if exc_type is not None:
                self._handle_error(exc_val)
            self._end_session_span()
        if hasattr(self._connection, "__aexit__"):
            return await self._connection.__aexit__(exc_type, exc_val, exc_tb)
        await self._connection.close()
        return False

    @dont_throw
    def _process_outgoing_event(self, event):
        """Process outgoing client events."""
        if isinstance(event, dict):
            event_type = event.get("type")
        else:
            event_type = getattr(event, "type", None)

        if event_type == "response.create":
            self._processor.start_response_span(end_existing=True, set_input=True)
        elif event_type == "conversation.item.create":
            self._track_input(event)

    @dont_throw
    def _track_input(self, event):
        """Track input from conversation.item.create events."""
        if isinstance(event, dict):
            item = event.get("item", {})
        else:
            item = getattr(event, "item", {})

        if isinstance(item, dict):
            content = item.get("content", [])
            for block in content:
                if isinstance(block, dict):
                    if block.get("type") == "input_text":
                        self._state.input_text = block.get("text", "")
                    elif block.get("type") == "input_audio":
                        # Audio input - could capture transcript if available
                        pass

    def _handle_error(self, error):
        """Handle connection errors."""
        if self._state.response_span and self._state.response_span.is_recording():
            self._state.response_span.set_status(
                Status(StatusCode.ERROR, str(error))
            )
            self._state.response_span.record_exception(error)
            self._state.response_span.end()
            self._processor.reset_response_state()

        if self._state.session_span and self._state.session_span.is_recording():
            self._state.session_span.set_status(
                Status(StatusCode.ERROR, str(error))
            )
            self._state.session_span.record_exception(error)

    def _end_session_span(self):
        """End the session span."""
        if self._state.session_span and self._state.session_span.is_recording():
            self._state.session_span.set_status(Status(StatusCode.OK))
            self._state.session_span.end()


class RealtimeEventIterator:
    """Async iterator that wraps connection events and processes them."""

    def __init__(
        self,
        connection,
        state: RealtimeSessionState,
        processor: RealtimeEventProcessor,
    ):
        self._connection = connection
        self._state = state
        self._processor = processor
        self._iterator = connection.__aiter__()

    def __aiter__(self):
        return self

    async def __anext__(self):
        event = await self._iterator.__anext__()
        self._process_event(event)
        return event

    @dont_throw
    def _process_event(self, event):
        """Process server events to track response lifecycle."""
        event_type = getattr(event, "type", None)
        if not event_type:
            return

        # For response.created, start span if not already started
        if event_type == "response.created":
            self._processor.handle_response_created(event, start_span_if_none=True)
        else:
            # Delegate all other events to the shared processor
            self._processor.process_event(event)


class RealtimeSessionWrapper:
    """Wrapper for connection.session to track session updates."""

    def __init__(self, session, state: RealtimeSessionState):
        self._session = session
        self._state = state

    def __getattr__(self, name):
        return getattr(self._session, name)

    async def update(self, **kwargs):
        """Wrap session.update to track configuration changes."""
        result = await self._session.update(**kwargs)

        # Track session config updates
        if "session" in kwargs:
            session_config = kwargs["session"]
            if isinstance(session_config, dict):
                self._state.session_config.update(session_config)

                # Update session span attributes
                if self._state.session_span and self._state.session_span.is_recording():
                    if "temperature" in session_config:
                        _set_span_attribute(
                            self._state.session_span,
                            GenAIAttributes.GEN_AI_REQUEST_TEMPERATURE,
                            session_config["temperature"],
                        )
                    if "instructions" in session_config:
                        _set_span_attribute(
                            self._state.session_span,
                            GenAIAttributes.GEN_AI_SYSTEM_INSTRUCTIONS,
                            session_config["instructions"],
                        )

        return result


class RealtimeConversationWrapper:
    """Wrapper for connection.conversation to track input items."""

    def __init__(self, conversation, state: RealtimeSessionState):
        self._conversation = conversation
        self._state = state

    def __getattr__(self, name):
        return getattr(self._conversation, name)

    @property
    def item(self):
        """Return wrapped item accessor."""
        return RealtimeConversationItemWrapper(self._conversation.item, self._state)


class RealtimeConversationItemWrapper:
    """Wrapper for connection.conversation.item to track input."""

    def __init__(self, item, state: RealtimeSessionState):
        self._item = item
        self._state = state

    def __getattr__(self, name):
        return getattr(self._item, name)

    async def create(self, **kwargs):
        """Wrap item.create to track user input."""
        result = await self._item.create(**kwargs)

        # Track input for the span
        if "item" in kwargs:
            item = kwargs["item"]
            if isinstance(item, dict):
                content = item.get("content", [])
                for block in content:
                    if isinstance(block, dict):
                        if block.get("type") == "input_text":
                            self._state.input_text = block.get("text", "")

        return result


class RealtimeResponseWrapper:
    """Wrapper for connection.response to track response lifecycle."""

    def __init__(
        self,
        response,
        state: RealtimeSessionState,
        processor: RealtimeEventProcessor,
    ):
        self._response = response
        self._state = state
        self._processor = processor

    def __getattr__(self, name):
        return getattr(self._response, name)

    async def create(self, **kwargs):
        """Wrap response.create to start a response span."""
        # Start the response span before making the call
        self._processor.start_response_span(end_existing=True, set_input=True)

        return await self._response.create(**kwargs)

    async def cancel(self):
        """Wrap response.cancel to handle span cleanup."""
        result = await self._response.cancel()

        # End the response span with cancellation status
        if self._state.response_span and self._state.response_span.is_recording():
            self._state.response_span.set_status(
                Status(StatusCode.OK, "Response cancelled")
            )
            self._state.response_span.set_attribute("response.cancelled", True)
            self._state.response_span.end()
            self._processor.reset_response_state()

        return result


class RealtimeConnectionManagerWrapper:
    """
    Wrapper for the connection manager returned by client.beta.realtime.connect().

    This wraps the async context manager to:
    1. Start a session span when the connection is established
    2. Return a wrapped connection that instruments events
    3. End the session span when the connection closes
    """

    def __init__(self, connection_manager, tracer: Tracer, model: str):
        self._connection_manager = connection_manager
        self._tracer = tracer
        self._model = model
        self._state = None
        self._connection = None

    async def __aenter__(self):
        """Enter the connection manager and start session span."""
        if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
            return await self._connection_manager.__aenter__()

        # Start the session span
        self._state = RealtimeSessionState(self._tracer, self._model)
        parent_context = context_api.get_current()

        self._state.session_span = self._tracer.start_span(
            SPAN_NAME_SESSION,
            kind=SpanKind.CLIENT,
            context=parent_context,
        )

        # Set trace_context to include the session span as parent for response spans
        self._state.trace_context = trace.set_span_in_context(
            self._state.session_span, parent_context
        )

        if self._state.session_span.is_recording():
            _set_span_attribute(
                self._state.session_span,
                GenAIAttributes.GEN_AI_SYSTEM,
                "openai",
            )
            _set_span_attribute(
                self._state.session_span,
                GenAIAttributes.GEN_AI_REQUEST_MODEL,
                self._model,
            )
            _set_span_attribute(
                self._state.session_span,
                SpanAttributes.LLM_REQUEST_TYPE,
                "realtime",
            )

        # Enter the underlying connection manager
        connection = await self._connection_manager.__aenter__()

        # Wrap the connection
        self._connection = RealtimeConnectionWrapper(connection, self._state)
        return self._connection

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Exit the connection manager and end session span."""
        if self._state is None:
            return await self._connection_manager.__aexit__(exc_type, exc_val, exc_tb)

        try:
            # Handle any errors
            if exc_type is not None and self._state.session_span:
                if self._state.session_span.is_recording():
                    self._state.session_span.set_status(
                        Status(StatusCode.ERROR, str(exc_val))
                    )
                    self._state.session_span.record_exception(exc_val)
            elif self._state.session_span and self._state.session_span.is_recording():
                self._state.session_span.set_status(Status(StatusCode.OK))

            # End the session span
            if self._state.session_span:
                self._state.session_span.end()

        finally:
            # Exit the underlying connection manager
            # Don't return the result - let any exception propagate naturally
            await self._connection_manager.__aexit__(exc_type, exc_val, exc_tb)


@_with_tracer_wrapper
def realtime_connect_wrapper(tracer: Tracer, wrapped, instance, args, kwargs):
    """
    Wrapper for client.beta.realtime.connect().

    Returns a wrapped connection manager that instruments the session.
    """
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    model = kwargs.get("model", "gpt-4o-realtime-preview")
    connection_manager = wrapped(*args, **kwargs)
    return RealtimeConnectionManagerWrapper(connection_manager, tracer, model)



"""
Policies sub-client for Overmind API.
"""

from typing import Any, Dict, List, Optional, Union

from .models import PolicyCreateRequest, PolicyResponse, PolicyUpdateRequest


class PoliciesClient:
    """
    Sub-client for managing policies in the Overmind API.
    """

    def __init__(self, parent_client):
        self._client = parent_client

    def create(
        self,
        policy_id: str,
        policy_template: str,
        policy_description: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        is_input_policy: Optional[bool] = True,
        is_output_policy: Optional[bool] = True,
        stats: Optional[Dict[str, Any]] = None,
        *,
        policy_data: Optional[Union[PolicyCreateRequest, Dict[str, Any]]] = None,
    ) -> PolicyResponse:
        """
        Create a new policy.

        Args:
            policy_id: Unique identifier for the policy
            policy_description: Description of the policy
            parameters: Policy parameters
            policy_template: Policy template to use
            is_input_policy: Whether this is an input policy
            is_output_policy: Whether this is an output policy
            stats: Policy statistics
            policy_data: Alternative: pass a complete PolicyCreateRequest or dict (for backward compatibility)
        """
        if policy_data is not None:
            if isinstance(policy_data, dict):
                policy_data = PolicyCreateRequest(**policy_data)
            elif isinstance(policy_data, PolicyCreateRequest):
                pass
            else:
                raise TypeError("policy_data must be a dict or PolicyCreateRequest")
        else:
            policy_data = PolicyCreateRequest(
                policy_id=policy_id,
                policy_description=policy_description,
                parameters=parameters,
                policy_template=policy_template,
                is_input_policy=is_input_policy,
                is_output_policy=is_output_policy,
                stats=stats or {},
            )

        response_data = self._client._make_request(
            "PUT", "policies/add_policy", data=policy_data.model_dump()
        )
        return response_data

    def list(self, policy_type: Optional[str] = None) -> List[PolicyResponse]:
        """
        List all policies with optional filtering by type.

        Args:
            policy_type: Optional filter to show only input or output policies
        """
        params = {"policy_type": policy_type} if policy_type else None
        response_data = self._client._make_request(
            "GET", "policies/list_policies", params=params
        )
        return [PolicyResponse(**policy) for policy in response_data]

    def get(self, policy_id: str) -> PolicyResponse:
        """
        Get a specific policy by ID.

        Args:
            policy_id: The unique identifier of the policy to retrieve
        """
        response_data = self._client._make_request("GET", f"policies/view/{policy_id}")
        return PolicyResponse(**response_data)

    def update(
        self,
        policy_id: str,
        policy_template: str,
        policy_description: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        is_input_policy: Optional[bool] = None,
        is_output_policy: Optional[bool] = None,
        stats: Optional[Dict[str, Any]] = None,
        *,
        policy_data: Optional[Union[PolicyUpdateRequest, Dict[str, Any]]] = None,
    ) -> Dict[str, str]:
        """
        Update an existing policy.

        Args:
            policy_id: Unique identifier for the policy
            policy_description: Description of the policy
            parameters: Policy parameters
            engine: Policy engine
            is_input_policy: Whether this is an input policy
            is_output_policy: Whether this is an output policy
            stats: Policy statistics
            policy_data: Alternative: pass a complete PolicyUpdateRequest or dict (for backward compatibility)
        """
        if policy_data is not None:
            if isinstance(policy_data, dict):
                policy_data = PolicyUpdateRequest(**policy_data)
            elif isinstance(policy_data, PolicyUpdateRequest):
                pass
            else:
                raise TypeError("policy_data must be a dict or PolicyUpdateRequest")
        else:
            policy_data = PolicyUpdateRequest(
                policy_id=policy_id,
                policy_description=policy_description,
                parameters=parameters,
                policy_template=policy_template,
                is_input_policy=is_input_policy,
                is_output_policy=is_output_policy,
                stats=stats,
            )

        return self._client._make_request(
            "POST", "policies/edit_policy", data=policy_data.model_dump()
        )

    def delete(self, policy_id: str) -> Dict[str, str]:
        """
        Delete a policy by ID.

        Args:
            policy_id: The unique identifier of the policy to delete
        """
        return self._client._make_request("GET", f"policies/delete/{policy_id}")
